#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

install_requires = \
['get_version']

extras_require = \
{'test': ['pytest', 'pytest-black']}

setup(name='allot',
      version='0.1',
      description='A more fine-grained functools.singledispatch',
      author='Philipp A.',
      author_email='flying-sheep@web.de',
      url='https://github.com/flying-sheep/allot',
      py_modules=['allot'],
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
